#!/bin/ksh

XARGS=/usr/bin/xargs
LS=/bin/ls
ECHO=/bin/echo
DATE=/bin/date
CAT=/bin/cat
TEE=/usr/bin/tee
WC=/usr/bin/wc
CP=/bin/cp
RM=/bin/rm
TAR=/bin/tar
CHOWN=/bin/chown
CHGRP=/bin/chgrp
CHMOD=/bin/chmod
MKDIR=/bin/mkdir
TOUCH=/bin/touch
UNAME=/bin/uname
FIND=/usr/bin/find
GZIP=/usr/bin/gzip
GUNZIP=/usr/bin/gunzip

OS=`/bin/uname`
case $OS in
   AIX )
      AWK=/usr/bin/awk
      ;;
   Linux )
      AWK=/usr/bin/awk
      ;;
   SunOS )
      AWK=/usr/bin/nawk
      ;;
esac

usage()
{
${ECHO} "Usage: $0 DataMart-install-dir
DataMart-install-dir is the directory where DataMart is installed
"
	return 0
}

logError()
{
    ${ECHO} "Error: " "$*"
}

logAction()
{
   LOG_FILE=$1
   shift

   if [ "${LOG_FILE}" != "" ] && [ -f ${LOG_FILE} ]
   then
      ${ECHO} "$*" | ${TEE} -a ${LOG_FILE}
   else
      ${ECHO} "$*"
   fi
}

unpack_files()
{
    [ ! -d ${PATCH_ROOT} ] && logError "Missing patch root directory ${PATCH_ROOT}" && return 1
    [ ! -f ${PATCH_ARC} ] && logError "Missing patch archive ${PATCH_ARC}" && return 1

	${MKDIR} -p ${PATCH_DIR}
	[ ! -d ${PATCH_DIR} ] && logError "Could not make directory ${PATCH_DIR}" && return 1
	
	cd ${PATCH_DIR}
	gunzip -c ${PATCH_ARC} | ${TAR} xf - && return 0
	logError "problem unpacking ${PATCH_ARC}" && return 1
}

install_files()
{
    result=0
    cd ${PATCH_DIR}

    MANIFEST=${PATCH_DIR}/manifest.txt
    PATCH_ID=${PATCH_DIR}/patchID

    [ ! -f ${MANIFEST} ] && logError "Missing manifest file" && return 1
    [ ! -f ${PATCH_ID} ] && logError "Missing patchID file" && return 1
    [ ! -f ${DM_ENV} ] && logError "Missing ${DM_ENV} file" && return 1

    patchID="`${CAT} ${PATCH_ID} | ${XARGS}`"
    patchItems=`${CAT} ${MANIFEST} | ${WC} -l | ${XARGS}`
    patchRollback=${DM_PATCHES}/${patchID}.rollback.sh
    patchBackup=${DM_PATCHES}/${patchID}.rollback.tar
    patchExtract=${PATCH_DIR}/patch

    if [ ! -d ${patchExtract} ]
    then
       logError "Missing patch extract directory: ${patchExtract}"
       return 1
    fi

    dmUID=`${LS} -l ${DM_ENV} | ${XARGS} | ${AWK} '{print $3}'`
    dmGID=`${LS} -l ${DM_ENV} | ${XARGS} | ${AWK} '{print $4}'`

    if [ ! -d ${DM_PATCHES} ]
    then
       ${MKDIR} -p ${DM_PATCHES}
    fi

    if [ ! -d ${DM_PATCHES} ]
    then
       logError "Missing patch directory: ${DM_PATCHES}"
       return 1
    fi
    ${TOUCH} ${DM_PATCHES}/${patchID}.mdv

    # set owner, permission on DM patch dir
    ${CHOWN} ${dmUID} ${DM_PATCHES}
    ${CHGRP} ${dmGID} ${DM_PATCHES}
    ${CHMOD} 755 ${DM_PATCHES}

    # initialize the log file
    PATCH_LOG=${DM_PATCHES}/${patchID}-${TIMESTAMP}.log
    if [ ! -f ${PATCH_LOG} ]
    then
        ${TOUCH} ${PATCH_LOG}
    fi

    # initialize the rollback script
    if [ -f ${patchRollback} ]
    then
       ${RM} -f ${patchRollback}
    fi
    ${TOUCH} ${patchRollback}
    ${CHMOD} 755 ${patchRollback}
    ${ECHO} "#!/bin/ksh" >> ${patchRollback}
    ${ECHO} "${ECHO} \"Uninstalling patch: ${patchID}\"" >> ${patchRollback}
    ${ECHO} "if [ ! -f ${patchBackup}.gz ]" >> ${patchRollback}
    ${ECHO} "then" >> ${patchRollback}
    ${ECHO} "   ${ECHO} \"Error - cannot restore, patch backup file missing: ${patchBackup}.gz\"" >> ${patchRollback}
    ${ECHO} "   exit 1" >> ${patchRollback}
    ${ECHO} "fi" >> ${patchRollback}
    ${ECHO} "cd ${DM_DIR}" >> ${patchRollback}
    ${ECHO} "${ECHO} \"Expanding patch backup: ${patchBackup}.gz\"" >> ${patchRollback}
    ${ECHO} "${GUNZIP} -c ${patchBackup}.gz | ${TAR} xvf -" >> ${patchRollback}

    # get current DM version info
    if [ -f ${DM_DIR}/version ]
    then
       DM_VERSION="`${CAT} ${DM_DIR}/version | ${XARGS}`"
    else
       DM_VERSION="(unknown)"
    fi

    # log preamble...
    logAction "${PATCH_LOG}" "--------------------------------------------------"
    logAction "${PATCH_LOG}" "Starting:     `${DATE} -u`"
    logAction "${PATCH_LOG}" "HostInfo:     `${UNAME} -a`"
    logAction "${PATCH_LOG}" "RunningAs:    `/usr/bin/id`"
    logAction "${PATCH_LOG}" "UserDetails:  uid=${dmUID} gid=${dmGID}"
    logAction "${PATCH_LOG}" "PatchID:      ${patchID}"
    logAction "${PATCH_LOG}" "NbItems:      ${patchItems}"
    logAction "${PATCH_LOG}" "PatchRoot:    ${PATCH_ROOT}"
    logAction "${PATCH_LOG}" "PatchDIR:     ${PATCH_DIR}"
    logAction "${PATCH_LOG}" "PatchExtract: ${patchExtract}"
    logAction "${PATCH_LOG}" "DataMart:     ${DM_DIR}"
    logAction "${PATCH_LOG}" "Version:      ${DM_VERSION}"

    # apply patch...
    for patchItem in `${CAT} ${MANIFEST} | ${XARGS}`
    do
       fileMode="`${ECHO} ${patchItem} | ${AWK} -F: '{print $1}'`"
       fileName="`${ECHO} ${patchItem} | ${AWK} -F: '{print $2}'`"
       logAction "${PATCH_LOG}" "PatchItem:    mode=${fileMode} file=${fileName}"
       patchSource=${patchExtract}/${fileName}
       dmTarget=${DM_DIR}/${fileName}
       if [ -f ${dmTarget} ]
       then
          dmBackup=${dmTarget}-${patchID}
          logAction "${PATCH_LOG}" "Backup:       ${dmTarget} to ${dmBackup}"
          ${CP} -fp ${dmTarget} ${dmBackup}
          ${CHOWN} ${dmUID} ${dmBackup}
          ${CHGRP} ${dmGID} ${dmBackup}
          ${CHMOD} ${fileMode} ${dmBackup}
          # write rollback
          ${ECHO} "${ECHO} \"Restore ${dmTarget} from ${dmBackup}\"" >> ${patchRollback}
          ${ECHO} "${CP} -fp ${dmBackup} ${dmTarget}" >> ${patchRollback}
          ${ECHO} "${ECHO} \"Set file owner ${dmUID}:${dmGID} and mode ${fileMode} on ${dmTarget}\"" >> ${patchRollback}
          ${ECHO} "${CHOWN} ${dmUID} ${dmTarget}" >> ${patchRollback}
          ${ECHO} "${CHGRP} ${dmGID} ${dmTarget}" >> ${patchRollback}
          ${ECHO} "${CHMOD} ${fileMode} ${dmTarget}" >> ${patchRollback}
       else
          # this is a new file
          logAction "${PATCH_LOG}" "NewFile:      ${patchSource} to ${dmTarget}"
          ${ECHO} "${ECHO} \"Remove new file ${dmTarget}\"" >> ${patchRollback}
          ${ECHO} "${RM} -f ${dmTarget}" >> ${patchRollback}
       fi
       if [ -f ${patchSource} ]
       then
          logAction "${PATCH_LOG}" "Install:      ${patchSource} to ${dmTarget}"
          ${CP} -f ${patchSource} ${dmTarget}
          logAction "${PATCH_LOG}" "SetOwner:     ${dmUID}:${dmGID} ${dmTarget}"
          ${CHOWN} ${dmUID} ${dmTarget}
          ${CHGRP} ${dmGID} ${dmTarget}
          logAction "${PATCH_LOG}" "SetMode:      ${fileMode} ${dmTarget}"
          ${CHMOD} ${fileMode} ${dmTarget}
       else
          logError "============================================================"
          logError "============================================================"
          logError ""
          logError "Missing patch file: ${fileName} (${patchSource})"
          logError "MDV may be partially applied, recommend executing rollback script"
          logError "As root, run this script: ${patchRollback}"
          logError ""
          logError "============================================================"
          logError "============================================================"
          result=1
          break
       fi
    done

    # remove any prior backup for this patch
    ${RM} -f ${patchBackup}*

    # now create the rollback backup file
    cd ${DM_DIR}
    logAction "${PATCH_LOG}" "Archive:      Creating restore archive..."
    ${TAR} cvf ${patchBackup} ${DM_PATCHES}/${patchID}.mdv
    # find files coped to backup versions, with patchID as extension, add to tar
    ${FIND} . -name \*-${patchID} -ls -exec ${TAR} rvf ${patchBackup} {} \;
    ${GZIP} ${patchBackup}

    # and delete the replaced files... 
    cd ${DM_DIR}
    logAction "${PATCH_LOG}" "CleanUp:      Removing replaced files..."
    ${FIND} . -name \*-${patchID} -ls -exec ${RM} -f {} \;

    # cleanup the files from the tar
    ${ECHO} "${ECHO} \"Removing restore files...\"" >> ${patchRollback}
    ${ECHO} "cd ${DM_DIR}" >> ${patchRollback}
    ${ECHO} "${FIND} . -name \\*-${patchID} -ls -exec ${RM} -f {} \\;" >> ${patchRollback}
    # one final comment in the rollback script
    ${ECHO} "${ECHO} \"Done\"" >> ${patchRollback}

    if [ ${result} -eq 0 ]
    then
       status="success"
    else
       status="error"
    fi
    logAction "${PATCH_LOG}" "Completed:    `date -u`"
    logAction "${PATCH_LOG}" "Status:       ${status}"
    logAction "${PATCH_LOG}" "--------------------------------------------------"

    return ${result}
}

cleanup_files()
{
    cd ${PATCH_ROOT}
	${RM} -fr ${PATCH_DIR} && return 0
	logError "Could not remove ${PATCH_DIR}" && return 1
}

# ----------------------------------------
#
# main()
#

if [ $#  -ne 1 ]
then
  usage
  exit 1
fi

TIMESTAMP=`${DATE} -u +"%Y.%m.%d-%H.%M.%S"-UTC`

PATCH_ROOT=`pwd`
PATCH_PATH=MDV
PATCH_FILE=dm.tar.gz
PATCH_DIR=${PATCH_ROOT}/${PATCH_PATH}
PATCH_ARC=${PATCH_ROOT}/${PATCH_FILE}

DM_DIR="$1"
DM_ENV=${DM_DIR}/dataMart.env
DM_PATCHES=${DM_DIR}/patches

if [ ! -f ${DM_ENV} ]
then
   logError "Cannot locate ${DM_ENV}"
   exit 1
fi

echo "Unpacking..." && unpack_files || exit 1
echo "Installing..." && install_files || exit 1
echo "Cleaning up..." && cleanup_files || exit 1
echo "Success."
echo "Install complete"
exit 0
